/* @(#) client/server context implementation
 *
 * Copyright 2008 Pavel V. Cherenkov (pcherenkov@gmail.com)
 *
 *  This file is part of udpxy.
 *
 *  udpxy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  udpxy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with udpxy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "osdef.h"  /* os-specific definitions */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/in.h>

#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include "ctx.h"
#include "udpxy.h"
#include "util.h"
#include "mtrace.h"

extern FILE* g_flog;

/* initialize server context data
 */
int
init_server_ctx( struct server_ctx* ctx,
                 const size_t max,
                 const char* laddr, uint16_t lport,
                 const char* mifc_addr )
{
    int flags = -1;

    assert( lport && mifc_addr && ctx && max );

    ctx->lsockfd = 0;
    (void) strncpy( ctx->listen_addr, (laddr ? laddr : IPv4_ALL),
                    IPADDR_STR_SIZE );
    ctx->listen_addr[ IPADDR_STR_SIZE - 1 ] = '\0';

    ctx->listen_port = lport;

    (void) strncpy( ctx->mcast_ifc_addr, mifc_addr, IPADDR_STR_SIZE );
    ctx->mcast_ifc_addr[ IPADDR_STR_SIZE - 1 ] = '\0';

    ctx->cl = calloc(max, sizeof(struct client_ctx));
    if( NULL == ctx->cl ) {
        mperror( g_flog, errno, "client_t - calloc");
        return ERR_INTERNAL;
    }

    (void) memset( ctx->cl, 0, max * sizeof(struct client_ctx) );
    ctx->clfree = ctx->clmax = max;

    if( 0 != pipe(ctx->cpipe) ) {
        mperror( g_flog, errno, "init_server_ctx - pipe" );
        return ERR_INTERNAL;
    }

    /* make reading end of pipe non-blocking (we don't want to
     * block on pipe read on the server side)
     */
    if( -1 == (flags = fcntl( ctx->cpipe[0], F_GETFL )) ||
        -1 == fcntl( ctx->cpipe[0], F_SETFL, flags | O_NONBLOCK ) ) {
        mperror( g_flog, errno, "init_server_ctx - fcntl" );
        return ERR_INTERNAL;
    }

    return 0;
}


void
free_server_ctx( struct server_ctx* ctx )
{
    int i = -1;

    assert(ctx);
    free( ctx->cl );

    for( i = 0; i < 2; ++i ) {
        if( ctx->cpipe[i] < 0 ) continue;

        if( 0 != close( ctx->cpipe[i] ) ) {
            mperror( g_flog, errno,
                    "free_server_ctx - close(pipe)" );
        }
    }

    (void) memset( ctx, 0, sizeof(struct server_ctx) );
    return;
}


/* find index of the first client with the given pid
 */
int
find_client( const struct server_ctx* ctx, pid_t pid )
{
    int i = -1;

    assert( ctx && (pid >= 0) );

    for( i = 0; (size_t)i < ctx->clmax; ++i ) {
        if( ctx->cl[ i ].pid == pid )
            return i;
    }

    return -1;
}

/* populate connection's source info in client context
 */
static int
get_src_info( struct client_ctx* cl, int sockfd )
{
    struct stat st;
    struct sockaddr_in addr;
    a_socklen_t len = 0;
    int rc = 0;

    assert( cl );
    if( -1 == (rc = fstat( sockfd, &st )) ) {
        mperror( g_flog, errno, "get_src_info - fstat" );
        return ERR_INTERNAL;
    }

    if( S_ISREG( st.st_mode ) ) {
        (void) strncpy( cl->src_addr, "File", sizeof(cl->src_addr) );
        cl->src_addr[ sizeof(cl->src_addr) - 1 ] = '\0';
        cl->src_port = 0;
    }
    else if( S_ISSOCK( st.st_mode ) ) {
        len = sizeof(addr);
        rc = getpeername( sockfd, (struct sockaddr*)&addr, &len );
        if( 0 == rc ) {
            (void)strncpy( cl->src_addr, inet_ntoa(addr.sin_addr),
                    sizeof(cl->src_addr) - 1 );
            cl->src_addr[ sizeof(cl->src_addr) - 1 ] = '\0';

            cl->src_port = ntohs(addr.sin_port);
        }
        else {
            mperror( g_flog, errno, "get_src_info - getpeername");
            return ERR_INTERNAL;
        }
    } /* S_ISSOCK */

    return rc;
}


/* add client to server context
 */
int
add_client( struct server_ctx* ctx,
            pid_t cpid, const char* maddr, uint16_t mport,
            int sockfd )
{
    struct client_ctx* client = NULL;
    int index = -1;
    int rc = 0;

    assert( ctx && maddr && mport );

    index = find_client( ctx, 0 );
    if( -1 == index )
        return ERR_INTERNAL;

    client = &(ctx->cl[ index ]);
    client->pid = cpid;

    (void) strncpy( client->mcast_addr, maddr, IPADDR_STR_SIZE );
    client->mcast_addr[ IPADDR_STR_SIZE - 1 ] = '\0';

    client->mcast_port = mport;

    rc = get_src_info( client, sockfd );
    if( 0 != rc ) {
        client->pid = 0;
        return ERR_INTERNAL;
    }

    ctx->clfree--;

    if( g_flog ) {
        (void)fprintf( g_flog, "Added client: pid=[%d], maddr=[%s], mport=[%d], "
                "saddr=[%s], sport=[%d]\n",
                (int)cpid, maddr, (int)mport, client->src_addr, client->src_port );
    }
    return 0;
}


/* delete client from server context
 */
int
delete_client( struct server_ctx* ctx, pid_t cpid )
{
    struct client_ctx* client = NULL;
    int index = -1;

    assert( ctx && (cpid > 0) );

    index = find_client( ctx, cpid );
    if( -1 == index ) {
        return ERR_INTERNAL;
    }

    client = &(ctx->cl[ index ]);
    client->pid = 0;

    ctx->clfree++;

    if( g_flog ) {
        TRACE( (void)fprintf( g_flog, "Deleted client: pid=[%d]\n", cpid) );
    }

    return 0;
}

/* reset packet-buffer registry in stream spec
 */
void
reset_pkt_registry( struct dstream_ctx* ds )
{
    assert( ds );

    ds->flags &= ~F_DROP_PACKET;
    ds->pkt_count = 0;
}


/* release resources allocated for stream spec
 */
void
free_dstream_ctx( struct dstream_ctx* ds )
{
    assert( ds );

    if( NULL != ds->pkt ) {
        free( ds->pkt );
        ds->pkt = NULL;

        ds->pkt_count = ds->max_pkt = 0;
    }
}

/* __EOF__ */

