/* @(#) udpxy server: main module
 *
 * Copyright 2008 Pavel V. Cherenkov (pcherenkov@gmail.com)
 *
 *  This file is part of udpxy.
 *
 *  udpxy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  udpxy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with udpxy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "osdef.h"  /* os-specific definitions */

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include <netinet/in.h>
#include <sys/select.h>

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <fcntl.h>
#include <syslog.h>
#include <time.h>
#include <sys/uio.h>

#include "mtrace.h"
#include "rparse.h"
#include "util.h"
#include "prbuf.h"
#include "ifaddr.h"

#include "udpxy.h"
#include "ctx.h"
#include "mkpg.h"
#include "rtp.h"
#include "uopt.h"
#include "dpkt.h"

/* socket timeouts in seconds */
static const u_short RLY_SOCK_TIMEOUT   = 5;
static const u_short SRVSOCK_TIMEOUT    = 1;

static const int LQ_BACKLOG = 1;    /* server backlog value */

/* server commands
 */
static const char  CMD_UDP[]            = "udp";
static const char  CMD_STATUS[]         = "status";
static const char  CMD_RESTART[]        = "restart";
static const char  CMD_RTP[]            = "rtp";

/* globals
 */

FILE*               g_flog      = NULL;
struct udpxy_opt    g_uopt;

/* misc
 */

static volatile sig_atomic_t g_quit = 0;
static volatile sig_atomic_t g_childexit = 0;


/*********************************************************/

/* handler for signals requestin application exit
 */
static void
handle_quitsigs(int signo)
{
    g_quit = (sig_atomic_t)1;
    (void) &signo;

    TRACE( (void)fprintf(g_flog,
                "*** Caught SIGNAL %d in process=[%d] ***\n",
                signo, getpid()) );
    return;
}


/* return 1 if the application must gracefully quit
 */
static sig_atomic_t
must_quit() { return g_quit; }


/* handle SIGCHLD
 */
static void
handle_sigchld(int signo)
{
    (void) &signo;
    g_childexit = 1;

    TRACE( (void)fprintf(g_flog, "*** Caught SIGCHLD in process=[%d] ***\n",
                getpid()) );
    return;
}


static int get_childexit()      { return g_childexit; }


/* clear SIGCHLD flag and adjust context if needed
 */
static void
wait_children( struct server_ctx* ctx, int options )
{
    int status;
    pid_t pid;

    assert( ctx );
    g_childexit = 0;

    while( 0 < (pid = waitpid( -1, &status, options )) ) {
        TRACE( (void)fprintf( g_flog, "Client [%d] has exited.\n", pid) );
        delete_client( ctx, pid );
    }

    if( (-1 == pid) && ( ECHILD != errno ) ) {
        mperror(g_flog, errno, "waitpid");
    }

    return;
}


/* wait for all children to quit
 */
static void
wait_all( struct server_ctx* ctx ) { wait_children( ctx, 0 ); }


/* wait for the children who already terminated
 */
static void
wait_terminated( struct server_ctx* ctx ) { wait_children( ctx, WNOHANG ); }


/* read HTTP request from sockfd, parse it into command
 * and its parameters (for instance, command='udp' and
 * parameters being '192.168.0.1:5002')
 */
static int
read_command( int sockfd, char* cmd, size_t clen,
              char* param, size_t plen )
{
#define DBUF_SZ 2048  /* max size for raw data with HTTP request */
#define RBUF_SZ 512   /* max size for url-derived request */
    static char httpbuf[ DBUF_SZ ],
                request[ RBUF_SZ ];
    ssize_t hlen;
    size_t  rlen;
    int rc = 0;

    assert( (sockfd > 0) && cmd && clen && param && plen );

    TRACE( (void)fprintf( g_flog,  "Reading command from socket [%d]\n", sockfd ) );
    hlen = recv( sockfd, httpbuf, sizeof(httpbuf), 0 );
    if( -1 == hlen ) {
        mperror(g_flog, errno, "read_command - recv");
        return errno;
    }

    /* deep DEBUG - re-enable if needed
    TRACE( (void) save_buffer( httpbuf, hlen, "/tmp/httpbuf.dat" ) );
    */

    rlen = sizeof(request);
    rc = get_request( httpbuf, (size_t)hlen, request, &rlen );
    if( -1 == rc ) return rc;

    TRACE( (void)fprintf( g_flog, "Request=[%s], length=[%lu]\n",
                request, (u_long)rlen ) );

    rc = parse_param( request, rlen, cmd, clen, param, plen );
    if( 0 == rc ) {
        TRACE( (void)fprintf( g_flog, "Command [%s] with params [%s]"
                    " read from socket=[%d]\n", cmd, param, sockfd) );
    }

    return rc;
}


/* terminate the client process
 */
static int
terminate( pid_t pid )
{
    TRACE( (void)fprintf( g_flog, "Forcing client process [%d] to QUIT\n", pid) );

    if( pid <= 0 ) return 0;

    if( 0 != kill( pid, SIGQUIT ) ) {
        if( ESRCH != errno ) {
            mperror(g_flog, errno, "terminate - kill");
            return ERR_INTERNAL;
        }
        /* ESRCH could mean client has quit already;
            * if so we should wait for it */

        TRACE( (void)fprintf( g_flog, "Process [%d] is not running.\n",
                pid ) );
    }

    return 0;
}


/*  terminate all clients
 */
static void
terminate_all_clients( struct server_ctx* ctx )
{
    size_t i;
    pid_t pid;

    for( i = 0; i < ctx->clmax; ++i ) {
        pid = ctx->cl[i].pid;
        if( pid > 0 ) (void) terminate( pid );
    }

    return;
}


/* send HTTP response to socket
 */
static int
send_http_response( int sockfd, int code, const char* reason )
{
    char    msg[ 128 ];
    ssize_t nsent;
    a_socklen_t msglen;

    assert( (sockfd > 0) && code && reason );

    msglen = snprintf( msg, sizeof(msg) - 1, "HTTP/1.0 %d %s \n\n",
              code, reason );
    if( msglen <= 0 ) return ERR_INTERNAL;

    nsent = send( sockfd, msg, msglen, 0 );
    if( -1 == nsent ) {
        mperror(g_flog, errno, "send_http_response - send");
        return ERR_INTERNAL;
    }

    TRACE( (void)fprintf( g_flog, "Sent HTTP response code=[%d], reason=[%s] to socket=[%d]\n",
                code, reason, sockfd) );
    return 0;
}


/* set up (server) listening sockfd
 */
static int
setup_listener( const char* ipaddr, int port, int* sockfd )
{
    int rc, lsock;
    struct sockaddr_in servaddr;
    const int ON = 1;

    assert( (port > 0) && sockfd && ipaddr );
    TRACE( (void)fprintf( g_flog, "Setting up listener for [%s:%d]\n",
                ipaddr[0] ? ipaddr : IPv4_ALL, port) );

    rc = ERR_INTERNAL;
    do {
        lsock = socket( AF_INET, SOCK_STREAM, 0 );
        if( -1 == lsock ) break;

        (void) memset( &servaddr, 0, sizeof(servaddr) );
        servaddr.sin_family = AF_INET;
        servaddr.sin_port = htons( (short)port );

        if( '\0' != ipaddr[0] ) {
            if( 1 != inet_aton(ipaddr, &servaddr.sin_addr) ) {
                TRACE( (void)fprintf( g_flog, "Invalid server IP: [%s]\n",
                        ipaddr) );

                rc = ERR_PARAM;
                break;
            }
        }
        else {
            servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
        }

        rc = setsockopt( lsock, SOL_SOCKET, SO_REUSEADDR,
                         &ON, sizeof(ON) );
        if( 0 != rc ) {
            mperror(g_flog, errno, "setsockopt SO_REUSEADDR");
            break;
        }

        rc = bind( lsock, (struct sockaddr*)&servaddr, sizeof(servaddr) );
        if( 0 != rc ) break;

        rc = listen( lsock, LQ_BACKLOG );
        if( 0 != rc ) break;

        rc = 0;
    } while(0);

    if( 0 != rc ) {
        if(errno)
            mperror(g_flog, errno, "socket/bind/listen error");
        if( lsock ) {
            (void) close( lsock );
        }
    }
    else {
        *sockfd = lsock;
        TRACE( (void)fprintf( g_flog, "Server socket=[%d] is set up\n", lsock) );
    }

    return rc;
}


/* set up the socket to receive multicast data
 *
 */
static int
setup_mcast_listener( struct sockaddr_in*   sa,
                      const struct in_addr* mifaddr,
                      int*                  mcastfd )
{
    int sockfd, rc;
    int ON = 1;
    struct ip_mreq mreq;
    int buflen = 1024 * g_uopt.rbuf_K;

    assert( sa && mifaddr && mcastfd );

    TRACE( (void)fprintf( g_flog, "Setting up multicast listener\n") );
    rc = ERR_INTERNAL;
    do {
        sockfd = socket( AF_INET, SOCK_DGRAM, 0 );
        if( -1 == sockfd ) {
            mperror(g_flog, errno, "socket");
            break;
        }

        if( buflen != 0 ) {
            rc = setsockopt( sockfd, SOL_SOCKET, SO_RCVBUF,
                             &buflen, sizeof(buflen) );
            if( 0 != rc ) {
                mperror(g_flog, errno, "setsockopt SO_RCVBUF");
                break;
            }
            TRACE( (void)fprintf(g_flog,
                    "multicast buffer size set to [%d] bytes\n",
                    buflen ) );
        }


        rc = setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR,
                         &ON, sizeof(ON) );
        if( 0 != rc ) {
            mperror(g_flog, errno, "setsockopt SO_REUSEADDR");
            break;
        }

        rc = bind( sockfd, (struct sockaddr*)sa, sizeof(*sa) );
        if( 0 != rc ) {
            mperror(g_flog, errno, "bind");
            break;
        }

        (void) memset( &mreq, 0, sizeof(mreq) );
        (void) memcpy( &mreq.imr_multiaddr, &(sa->sin_addr),
                sizeof(struct in_addr) );

        (void) memcpy( &mreq.imr_interface, mifaddr,
                sizeof(struct in_addr) );

        rc = setsockopt( sockfd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                &mreq, sizeof(mreq) );
        if( 0 != rc ) {
            mperror(g_flog, errno, "setsockopt IP_ADD_MEMBERSHIP");
            break;
        }
    } while(0);

    if( 0 == rc ) {
        *mcastfd = sockfd;
        TRACE( (void)fprintf( g_flog, "Mcast listener socket=[%d] set up\n", sockfd) );
    }
    else {
        (void)close(sockfd);
    }

    return rc;
}


/* unsubscribe from multicast and close the reader socket
 *
 */
static void
close_mcast_listener( int msockfd, const struct in_addr* mifaddr )
{
    struct ip_mreq mreq;

    assert( mifaddr );

    if( msockfd <= 0 ) return;

    (void) memset( &mreq, 0, sizeof(mreq) );
    (void) memcpy( &mreq.imr_interface, mifaddr,
                sizeof(struct in_addr) );

    (void) setsockopt( msockfd, IPPROTO_IP, IP_DROP_MEMBERSHIP,
                    &mreq, sizeof(mreq) );
    (void) close( msockfd );

    TRACE( (void)fprintf( g_flog, "Mcast listener socket=[%d] closed\n", msockfd) );
    return;
}


/* set send/receive timeouts on socket(s)
 */
static int
set_timeouts( int rsock, int ssock, u_short rcv_tmout, u_short snd_tmout )
{
    struct timeval rtv, stv;
    int rc = 0;

    rtv.tv_sec = rcv_tmout;
    rtv.tv_usec = 0;
    rc = setsockopt( rsock, SOL_SOCKET, SO_RCVTIMEO, &rtv, sizeof(rtv) );
    if( -1 == rc ) {
        mperror(g_flog, errno, "setsockopt - SO_RCVTIMEO");
        return ERR_INTERNAL;
    }

    stv.tv_sec = snd_tmout;
    stv.tv_usec = 0;
    rc = setsockopt( ssock, SOL_SOCKET, SO_SNDTIMEO, &stv, sizeof(stv) );
    if( -1 == rc ) {
        mperror(g_flog, errno, "setsockopt - SO_SNDTIMEO");
        return ERR_INTERNAL;
    }

    return rc;
}


/* statistics on traffic relay & data gathering
 */
struct tps_data {
    pid_t  pid;         /* our PID - cached */
    time_t tm_from;     /* last time update sent (successfully) */
    double niter;       /* number of iterations since last try  */
    double nbytes;      /* bytes transferred since last update  */
};


/* init traffic relay statistics
 */
static void
tpstat_init( struct tps_data* d, int setpid )
{
    assert( d );

    if( setpid )
        d->pid = getpid();

    d->tm_from = time(NULL);
    d->niter = 0;
    d->nbytes = 0;

    return;
}


/* send statistics update to server (if it's time)
 */
static void
tpstat_update( struct server_ctx* ctx,
               struct tps_data* d, ssize_t nbytes )
{
    static const double MAX_NOUPDATE_ITER = 1000.0;
    static const double MAX_SEC = 10.0;

    struct tput_stat ts;
    int writefd = -1, nwr = -1, rc = 0;
    const int DO_NOT_SET_PID = 0;
    double nsec;

    assert( ctx && d );

    d->nbytes += nbytes;

    nsec = difftime( time(NULL), d->tm_from );
    d->niter++;
    if( !((d->niter >= MAX_NOUPDATE_ITER) || (nsec >= MAX_SEC)) )
        return;

    /* attempt to send update to server */
    d->niter = 0;

    ts.sender_id = d->pid;
    ts.nbytes = d->nbytes;
    ts.nsec = nsec;

    writefd = ctx->cpipe[1];
    do {
        nwr = write( writefd, &ts, sizeof(ts) );
        if( nwr <= 0 ) {
            if( (EINTR != errno) && (EAGAIN != errno) ) {
                mperror( g_flog, errno, "tpstat_update - write" );
                rc = -1;
                break;
            }
            /* if it's an interrupt or pipe full - ignore */
            TRACE( (void)fprintf( g_flog, "%s - write error [%s] - ingored\n",
                            __func__, strerror(errno)) );
            break;
        }
        if( sizeof(ts) != (size_t)nwr ) {
            (void)fprintf( g_flog, "%s - wrote [%d] bytes to pipe, expected [%u]\n",
                    __func__, nwr, (int)sizeof(ts) );
            rc = -1;
            break;
        }

        TRACE( (void)fprintf( g_flog,
                "Sent TSTAT={ sender=[%ld], bytes=[%f], seconds=[%f] }\n",
                (long)ts.sender_id, ts.nbytes, ts.nsec) );
    } while(0);


    if( 0 == rc ) {
        tpstat_init( d, DO_NOT_SET_PID );
    }

    return;
}


/* register received packet into registry (for scattered output)
 */
static int
register_packet( struct dstream_ctx* spc, char* buf, size_t len )
{
    struct iovec* new_pkt = NULL;
    static const int DO_VERIFY = 1;

    void* new_buf = buf;
    size_t new_len = len;

    assert( spc->max_pkt > 0 );

    /* enlarge packet registry if needed */
    if( spc->pkt_count >= spc->max_pkt ) {
        spc->max_pkt <<= 1;
        spc->pkt = realloc( spc->pkt, spc->max_pkt * sizeof(spc->pkt[0]) );
        if( NULL == spc->pkt ) {
            mperror( g_flog, errno, "register_packet - realloc" );
            return -1;
        }

        TRACE( (void)fprintf( g_flog, "RTP packet registry "
                "expanded to [%lu] records\n", (u_long)spc->max_pkt ) );
    }

    /* put packet info into registry */

    new_pkt = &(spc->pkt[ spc->pkt_count ]);

    /*
    TRACE( (void)fprintf( stderr, "IN: packet [%lu]: buf=[%p], len=[%lu]\n",
                (u_long)spc->pkt_count, (void*)buf, (u_long)len ) );
    */

    if( 0 != RTP_process( &new_buf, &new_len, DO_VERIFY, g_flog ) ) {
        TRACE( (void)fputs("register packet: dropping\n", g_flog) );
        spc->flags |= F_DROP_PACKET;

        return 0;
    }

    new_pkt->iov_base = new_buf;
    new_pkt->iov_len = new_len;

    /*
    TRACE( (void)fprintf( stderr, "OUT: packet [%lu]: buf=[%p], len=[%lu]\n",
                (u_long)spc->pkt_count, new_pkt->iov_base, (u_long)new_pkt->iov_len ) );
    */

    spc->pkt_count++;
    return 0;
}


/* read data from source, determine underlying protocol
 * (if not already known); and process the packet
 * if needed (for RTP - register packet)
 *
 * return the number of octets read from the source
 * into the buffer
 */
static ssize_t
read_packet( struct dstream_ctx* spc, int fd, char* buf, size_t len )
{
    ssize_t n = -1;
    size_t chunk_len = len;

    assert( spc && buf && len );
    assert( fd > 0 );

    /* if *RAW* data specified - read AS IS
     * and exit */
    if( DT_RAW == spc->stype ) {
        return read_buf( fd, buf, len, g_flog );
    }

    /* if it is (or *could* be) RTP, read only MTU bytes
     */
    if( (spc->stype == DT_RTP_TS) || (spc->flags & F_CHECK_FMT) )
        chunk_len = spc->mtu;

    if( spc->flags & F_FILE_INPUT ) {
        assert( !buf_overrun( buf, len, 0, chunk_len, g_flog ) );
        n = read_frecord( fd, buf, chunk_len, &(spc->stype), g_flog );
        if( n <= 0 ) return n;
    }
    else {
        assert( !buf_overrun(buf, len, 0, chunk_len, g_flog) );
        n = read_buf( fd, buf, chunk_len, g_flog );
        if( n <= 0 ) return n;

        spc->stype = get_mstream_type( buf, n, g_flog );
    }

    if( spc->flags & F_CHECK_FMT ) {
        if( DT_RTP_TS == spc->stype ) {
            /* scattered data (to exclude RTP headers):
                * use packet registry */
            spc->flags |= F_SCATTERED;
        }
        else if( DT_TS == spc->stype ) {
            spc->flags &= ~F_SCATTERED;
        }
        else {
            spc->stype = DT_RAW;
            TRACE( (void)fputs( "Unrecognized stream type\n", g_flog ) );
        }

        TRACE( (void)fprintf( g_flog, "Established stream as [%s]\n",
               fmt2str( spc->stype ) ) );

        spc->flags &= ~F_CHECK_FMT;
    }

    if( spc->flags & F_SCATTERED )
        if( -1 == register_packet( spc, buf, n ) ) return -1;

    return n;
}


/* read data from source of specified type (UDP socket or otherwise);
 * read as many fragments as specified (max_frgs) into the buffer
 */
static ssize_t
read_data( struct dstream_ctx* spc, int fd, char* data,
           const ssize_t data_len, const int max_frgs )
{
    int m = 0;
    ssize_t n = 0, nrcv = -1;

    assert( spc && (data_len > 0) );


    for( m = 0, n = 0; ((m < 0) ? 1 : (m < max_frgs)); ++m ) {
        nrcv = read_packet( spc, fd, data + n, data_len - n );
        if( nrcv <= 0 ) {
            if( EAGAIN == errno ) {
                (void)fprintf(g_flog,
                        "Receive on socket/file [%d] timed out\n",
                        fd);
            }

            if( 0 == nrcv ) (void) fputs( "read_data - EOF\n", g_flog );
            else {
                mperror(g_flog, errno, "read_data - read/recv");
            }

            break;
        }

        if( spc->flags & F_DROP_PACKET ) {
            spc->flags &= ~F_DROP_PACKET;
            continue;
        }

        n += nrcv;
        if( n >= (data_len - nrcv) ) break;
    } /* for */

    if( (nrcv > 0) && !n ) {
        TRACE( (void)fprintf( g_flog, "read_data: no data to send "
                    "out of [%d] packets\n", m ) );
        return -1;
    }

    return (nrcv > 0) ? n : -1;
}


/* write data to destination(s)
 */
static ssize_t
write_data( const struct dstream_ctx* spc,
            const char* data,
            const ssize_t len,
            int fd )
{
    ssize_t n = 0;

    assert( spc && data && len );
    if( fd <= 0 ) return 0;

    if( spc->flags & F_SCATTERED ) {
        n = writev( fd, spc->pkt, spc->pkt_count );
        if( n <= 0 ) {
            if( EAGAIN == errno ) {
                (void)fprintf(g_flog, "Write on fd=[%d] timed out\n", fd);
            }
            mperror( g_flog, errno, "write_data - writev(file)" );
            return -1;
        }
    }
    else {
        n = write_buf( fd, data, len, g_flog );
    }

    return (n > 0) ? n : -1;
}


/* initialize incoming-stream context
 */
static int
init_dstream_ctx( struct dstream_ctx* ds, const char* cmd, const char* fname )
{
    assert( ds && cmd );

    ds->flags = 0;
    ds->pkt = NULL;
    ds->max_pkt = ds->pkt_count = 0;
    ds->mtu = ETHERNET_MTU;

    if( NULL != fname ) {
        ds->stype = DT_UNKNOWN;
        ds->flags |= (F_CHECK_FMT | F_FILE_INPUT);
        TRACE( (void)fputs( "File stream, RTP check enabled\n", g_flog ) );
    }
    else if( 0 == strncmp( cmd, CMD_UDP, sizeof(CMD_UDP) ) ) {
        ds->stype = DT_UNKNOWN;
        ds->flags |= F_CHECK_FMT;
        TRACE( (void)fputs( "UDP stream, RTP check enabled\n", g_flog ) );
    }
    else if( 0 == strncmp( cmd, CMD_RTP, sizeof(CMD_RTP) ) ) {
        ds->stype = DT_RTP_TS;
        ds->flags |= F_SCATTERED;
        TRACE( (void)fputs( "RTP (over UDP) stream assumed,"
                    " no checks\n", g_flog ) );
    }
    else {
        TRACE( (void)fprintf( g_flog, "init_dstream_ctx: "
                    "Irrelevant command [%s]\n", cmd) );
        return -1;
    }

    if( -1 == g_uopt.rbuf_msgs ) {
        g_uopt.rbuf_msgs = (g_uopt.rbuf_K * 1024) / ds->mtu;
        TRACE( (void)fprintf( g_flog, "Inbound buffer set to "
                "[%d] messages\n", g_uopt.rbuf_msgs ) );
    }

    ds->pkt = calloc( g_uopt.rbuf_msgs, sizeof(ds->pkt[0]) );
    if( NULL == ds->pkt ) {
        mperror( g_flog, errno, "init_dstream_ctx - calloc" );
        return -1;
    }

    ds->max_pkt = g_uopt.rbuf_msgs;
    return 0;
}


/* check and report a deviation in values: n_was and n_is are not supposed
 * to differ more than by delta
 */
static void
check_fragments( const char* action, ssize_t total, ssize_t n_was, ssize_t n_is,
                 ssize_t delta )
{
    if( (n_is < (n_was - delta)) || (n_is > (n_was + delta)) ) {
        (void)fprintf( g_flog,
                "%s [%ld] bytes out of [%ld], last=[%ld]\n",
                action, (long)n_is, (long)total, (long)n_was);
    }
}


/* adjust nice value if needed
 */
static int
set_nice( int val )
{
    if( 0 != val ) {
        errno = 0;
        (void) nice( val );
        if( 0 != errno ) {
            mperror( g_flog, errno, "relay_traffic - nice" );
            return -1;
        }
        else {
            TRACE( (void)fprintf(g_flog, "Nice value incremented by %d\n", val) );
        }
    }

    return 0;
}


/* set socket's send buffer value
 */
static int
set_sendbuf( int sockfd, const size_t len )
{
    size_t data_len = len;
    int rc = 0;

    rc = setsockopt( sockfd, SOL_SOCKET, SO_SNDBUF,
                    &data_len, sizeof(data_len) );
    if( 0 != rc ) {
        mperror(g_flog, errno, "setsockopt SO_SNDBUF");
        return -1;
    }
    else {
        TRACE( (void)fprintf(g_flog,
            "send buffer size set to [%u] bytes for socket [%d]\n",
            data_len, sockfd ) );
    }

    return rc;
}


/* relay traffic from source to destination socket
 *
 */
static int
relay_traffic( int ssockfd, int dsockfd, struct server_ctx* ctx,
               int dfilefd )
{
    volatile sig_atomic_t quit = 0;

    int rc = 0;
    ssize_t nrcv = 0, nsent = 0, nwr = 0,
            lrcv = 0, lsent = 0;
    char* data      = NULL;
    int data_len    = g_uopt.rbuf_K * 1024;

    /* permissible variation in data-packet size */
    static const ssize_t t_delta = 0x20;

    struct dstream_ctx ds;

    static const int SET_PID = 1;
    struct tps_data tps;

    assert( ctx );

    /* INIT
     */
    rc= init_dstream_ctx( &ds, ctx->cmd, g_uopt.srcfile );
    if( 0 != rc ) return -1;

    (void) set_nice( g_uopt.nice_incr );

    do {
        if( NULL == g_uopt.srcfile ) {
            rc = set_timeouts( ssockfd, dsockfd, ctx->rcv_tmout, ctx->snd_tmout );
            if( 0 != rc ) break;
        }

        if( dsockfd > 0 ) {
            (void) set_sendbuf( dsockfd, data_len );

            rc = send_http_response( dsockfd, 200, "OK" );
            if( 0 != rc ) break;
        }

        data = malloc(data_len);
        if( NULL == data ) {
            mperror( g_flog, errno, "relay_traffic - malloc" );
            break;
        }

        if( g_uopt.cl_tpstat )
            tpstat_init( &tps, SET_PID );
    } while(0);

    TRACE( (void)fprintf( g_flog, "Relaying traffic from socket[%d] to socket[%d], "
            "buffer size=[%d], rcache=[%d]\n", ssockfd, dsockfd,
            data_len, g_uopt.rbuf_msgs) );

    /* RELAY LOOP
     */
    while( (0 == rc) && !(quit = must_quit()) ) {
        nrcv = read_data( &ds, ssockfd, data, data_len, g_uopt.rbuf_msgs );
        if( -1 == nrcv ) break;

        TRACE( check_fragments( "received new", data_len, lrcv, nrcv, t_delta ) );
        lrcv = nrcv;

        if( dsockfd && (nrcv > 0) ) {
            nsent = write_data( &ds, data, nrcv, dsockfd );
            if( -1 == nsent ) break;

            TRACE( check_fragments("sent", nrcv, lsent, nsent, t_delta) );
            lsent = nsent;
        }

        if( dfilefd && (nrcv > 0) ) {
            nwr = write_data( &ds, data, nrcv, dfilefd );
            if( -1 == nwr )
                break;
            TRACE( check_fragments( "wrote to file",
                    nrcv, lsent, nsent, t_delta ) );
        }

        if( ds.flags & F_SCATTERED ) reset_pkt_registry( &ds );

        if( uf_TRUE == g_uopt.cl_tpstat )
            tpstat_update( ctx, &tps, nsent );

    } /* end of RELAY LOOP */

    /* CLEANUP
     */
    TRACE( (void)fprintf( g_flog, "Exited relay loop: received=[%ld], "
        "sent=[%ld], quit=[%ld]\n", (long)nrcv, (long)nsent, (long)quit ) );

    free_dstream_ctx( &ds );
    if( NULL != data ) free( data );

    if( 0 != (quit = must_quit()) ) {
        TRACE( (void)fprintf( g_flog, "Child process=[%d] must quit\n",
                    getpid()) );
    }

    return rc;
}


/* process command to relay udp traffic
 *
 */
static int
udp_relay( int sockfd, const char* param, size_t plen,
           const struct in_addr* mifaddr,
           struct server_ctx* ctx )
{
    char                mcast_addr[ IPADDR_STR_SIZE ];
    struct sockaddr_in  addr;

    uint16_t    port;
    pid_t       new_pid;
    int         rc = 0, flags;
    int         msockfd = -1, sfilefd = -1,
                dfilefd = -1, srcfd = -1;
    char        dfile_name[ MAXPATHLEN ];

    assert( (sockfd > 0) && param && plen && ctx );

    TRACE( (void)fprintf( g_flog, "udp_relay : new_socket=[%d] param=[%s]\n",
                        sockfd, param) );
    do {
        rc = parse_udprelay( param, plen, mcast_addr, IPADDR_STR_SIZE, &port );
        if( 0 != rc ) {
            (void) fprintf( g_flog, "Error [%d] parsing parameters [%s]\n",
                            rc, param );
            break;
        }

        if( 1 != inet_aton(mcast_addr, &addr.sin_addr) ) {
            (void) fprintf( g_flog, "Invalid address: [%s]\n", mcast_addr );
            rc = ERR_INTERNAL;
            break;
        }

        addr.sin_family = AF_INET;
        addr.sin_port = htons( (short)port );

    } while(0);

    if( 0 != rc ) {
        (void) send_http_response( sockfd, 500, "Service error" );
        return rc;
    }

    /* start the (new) process to relay traffic */

    if( 0 != (new_pid = fork()) ) {
        rc = add_client( ctx, new_pid, mcast_addr, port, sockfd );
        return rc; /* parent returns */
    }

    /* child process:
     */
    TRACE( (void)fprintf( g_flog, "Client process=[%d] started for socket=[%d]\n",
                getpid(), sockfd) );

    (void)close( ctx->lsockfd );

    /* close the reading end of the comm. pipe */
    (void)close( ctx->cpipe[0] );
    ctx->cpipe[0] = -1;

    do {
        /* make write end of pipe non-blocking (we don't want to
        * block on pipe write while relaying traffic)
        */
        if( -1 == (flags = fcntl( ctx->cpipe[1], F_GETFL )) ||
            -1 == fcntl( ctx->cpipe[1], F_SETFL, flags | O_NONBLOCK ) ) {
            mperror( g_flog, errno, "udp_relay - fcntl" );
            rc = -1;
            break;
        }

        if( NULL != g_uopt.dstfile ) {
            (void) snprintf( dfile_name, MAXPATHLEN - 1,
                    "%s.%d", g_uopt.dstfile, getpid() );
            dfilefd = creat( dfile_name, S_IRUSR | S_IWUSR | S_IRGRP );
            if( -1 == dfilefd ) {
                mperror( g_flog, errno, "upd_relay - g_uopt.dstfile open" );
                rc = -1;
                break;
            }

            TRACE( (void)fprintf( g_flog,
                        "Dest file [%s] opened as fd=[%d]\n",
                        dfile_name, dfilefd ) );
        }
        else dfilefd = -1;

        if( NULL != g_uopt.srcfile ) {
            sfilefd = open( g_uopt.srcfile, O_RDONLY | O_NOCTTY );
            if( -1 == sfilefd ) {
                mperror( g_flog, errno, "upd_relay - g_uopt.srcfile open" );
                rc = -1;
            }
            else {
                TRACE( (void) fprintf( g_flog, "Source file [%s] opened\n",
                            g_uopt.srcfile ) );
                srcfd = sfilefd;
            }
        }
        else {
            rc = setup_mcast_listener( &addr, mifaddr, &msockfd );
            srcfd = msockfd;
        }
        if( 0 != rc ) break;

        rc = relay_traffic( srcfd, sockfd, ctx, dfilefd );
        if( 0 != rc ) break;

    } while(0);

    if( msockfd > 0 ) {
        close_mcast_listener( msockfd, mifaddr );
    }
    if( sfilefd > 0 ) {
       (void) close( sfilefd );
       TRACE( (void) fprintf( g_flog, "Source file [%s] closed\n",
                            g_uopt.srcfile ) );
    }
    if( dfilefd > 0 ) {
       (void) close( dfilefd );
       TRACE( (void) fprintf( g_flog, "Dest file [%s] closed\n",
                            dfile_name ) );
    }

    if( 0 != rc ) {
        (void) send_http_response( sockfd, 500, "Service error" );
    }

    (void) close( sockfd );
    free_server_ctx( ctx );

    closelog();

    TRACE( (void)fprintf( g_flog, "Child process=[%d] exits with rc=[%d]\n",
                getpid(), rc) );

    if( g_flog && (stderr != g_flog) ) {
        (void) fclose(g_flog);
    }

    free_uopt( &g_uopt );

    rc = ( 0 != rc ) ? ERR_INTERNAL : rc;
    exit(rc);   /* child exits */

    return rc;
}


/* send server status as HTTP response to the given socket
 */
static int
report_status( int sockfd, const struct server_ctx* ctx, int options )
{
    static char buf[ 8192 ];
    int rc = 0;
    ssize_t n, nsent;
    size_t nlen = 0;

    assert( (sockfd > 0) && ctx );

    (void) memset( buf, 0, sizeof(buf) );

    nlen = sizeof(buf);
    rc = mk_status_page( ctx, buf, &nlen, options | MSO_HTTP_HEADER );

    for( n = nsent = 0; (0 == rc) && (nsent < (ssize_t)nlen);  ) {
        errno = 0;
        n = send( sockfd, buf, nlen, 0 );

        if( (-1 == n) && (EINTR != errno) ) {
            mperror(g_flog, errno, "report_status - send");
            rc = ERR_INTERNAL;
            break;
        }

        nsent += n;
    }

    if( 0 != rc ) {
        TRACE( (void)fprintf( g_flog, "Error generating status report\n" ) );
    }
    else {
        /* DEBUG only
        TRACE( (void)fprintf( g_flog, "Saved status buffer to file\n" ) );
        TRACE( (void)save_buffer(buf, nlen, "/tmp/status-udpxy.html") );
        */
    }

    return rc;
}


/* process command within a request
 */
static int
process_command( int new_sockfd, struct server_ctx* ctx,
                 const char* param, size_t plen )
{
    int rc = 0;
    const int STAT_OPTIONS = 0;
    const int RESTART_OPTIONS = MSO_SKIP_CLIENTS | MSO_RESTART;

    assert( (new_sockfd > 0) && ctx && param );

    if( 0 == strncmp( ctx->cmd, CMD_UDP, sizeof(ctx->cmd) ) ||
        0 == strncmp( ctx->cmd, CMD_RTP, sizeof(ctx->cmd) ) ) {
        if( ctx->clfree ) {
            rc = udp_relay( new_sockfd, param, plen,
                            &(ctx->mcast_inaddr), ctx );
        }
        else {
            send_http_response( new_sockfd, 401, "Bad request" );
            (void)fprintf( g_flog, "Client limit [%d] has been reached.\n",
                    ctx->clmax);
        }
    }
    else if( 0 == strncmp( ctx->cmd, CMD_STATUS, sizeof(ctx->cmd) ) ) {
        rc = report_status( new_sockfd, ctx, STAT_OPTIONS );
    }
    else if( 0 == strncmp( ctx->cmd, CMD_RESTART, sizeof(ctx->cmd) ) ) {
        (void) report_status( new_sockfd, ctx, RESTART_OPTIONS );

        terminate_all_clients( ctx );
        wait_all( ctx );
    }
    else {
        TRACE( (void)fprintf( g_flog, "Unrecognized command [%s] - ignoring.\n", ctx->cmd) );
        send_http_response( new_sockfd, 200, "OK" );
    }

    return rc;

}


/* read client statistics data and update the context
 */
static int
update_client_stat( struct server_ctx* ctx )
{
    int cindex = -1;
    int readfd = -1;
    ssize_t nread = 0;
    struct tput_stat ts;
    struct client_ctx* client = NULL;

    assert( ctx );

    readfd = ctx->cpipe[0];
    assert( readfd > 0 );

    (void)memset( &ts, 0, sizeof(ts) );

    nread = read( readfd, &ts, sizeof(ts) );
    if( nread <= 0 ) {
        if( (EINTR != errno) && (EAGAIN != errno) ) {
            mperror( g_flog, errno, "update_client_stat - read" );
            return ERR_INTERNAL;
        }
        /* if it's an interrupt or no data available - ignore */
        TRACE( (void)fprintf( g_flog, "%s - read error [%s] - ingored\n",
                        __func__, strerror(errno)) );
        return 0;
    }
    if( sizeof(ts) != (size_t)nread ) {
        (void)fprintf( g_flog, "%s - read [%d] bytes from pipe, expected [%u]\n",
                __func__, nread, (int)sizeof(ts) );
        return ERR_INTERNAL;
    }

    TRACE( (void)fprintf( g_flog,
            "Received TSTAT={ sender=[%ld], bytes=[%f], seconds=[%f] }\n",
            (long)ts.sender_id, ts.nbytes, ts.nsec) );

    cindex = find_client( ctx, (pid_t)ts.sender_id );
    if( -1 == cindex ) {
        (void)fprintf( g_flog, "%s - cannot find client [%ld]\n",
                __func__, (long)ts.sender_id );
        /* ignore invalid client id's */
        return 0;
    }

    client = &(ctx->cl[ cindex ]);
    client->tstat = ts;

    TRACE( (void)fprintf( g_flog, "Updated context for pid=[%d]; "
                "[%.1f] Kb/sec\n", client->pid, (ts.nbytes / 1024) / ts.nsec  ) );
    return 0;
}


/* process client requests
 */
static int
server_loop( const char* ipaddr, int port,
             const char* mcast_addr )
{
    int                 rc, maxfd,
                        new_sockfd;
    struct sockaddr_in  cliaddr;
    struct in_addr      mcast_inaddr;
    char                param[ 128 ];
    struct server_ctx   srv;
    a_socklen_t         addrlen;
    fd_set              rset;

    assert( (port > 0) && mcast_addr && ipaddr );

    (void)fprintf( g_flog, "Starting server [%d]; capacity=[%u] clients\n",
                        getpid(), g_uopt.max_clients );

    if( 1 != inet_aton(mcast_addr, &mcast_inaddr) ) {
        mperror(g_flog, errno, "inet_aton");
        return ERR_INTERNAL;
    }

    init_server_ctx( &srv, g_uopt.max_clients,
            (ipaddr[0] ? ipaddr : "0.0.0.0") , (uint16_t)port, mcast_addr );

    srv.rcv_tmout = srv.snd_tmout = RLY_SOCK_TIMEOUT;
    srv.mcast_inaddr = mcast_inaddr;

    if( 0 != (rc = setup_listener( ipaddr, port, &srv.lsockfd )) )
        return rc;

    new_sockfd = -1;
    TRACE( (void)fprintf( g_flog, "Entering server loop\n") );

    while( !must_quit() ) {
        if( get_childexit() ) {
            wait_terminated( &srv );
        }

        FD_ZERO( &rset );
        FD_SET( srv.lsockfd, &rset );
        FD_SET( srv.cpipe[0], &rset );

        TRACE( (void)fprintf( g_flog, "Server is waiting for input: socket=[%d], pipe=[%d]\n",
                    srv.lsockfd, srv.cpipe[0]) );

        maxfd = (srv.lsockfd > srv.cpipe[0] ) ? srv.lsockfd : srv.cpipe[0];
        rc = select( maxfd + 1, &rset, NULL, NULL, NULL );
        if( -1 == rc ) {
            if( EINTR == errno ) {
                if( must_quit() ) {
                    TRACE( (void)fprintf( g_flog, "\nServer must quit.\n") );
                    rc = 0;
                    break;
                }
                else
                    continue;
            }

            mperror( g_flog, errno, "select" );
            break;
        }

        if( FD_ISSET(srv.cpipe[0], &rset) ) {
            if( 0 != update_client_stat( &srv ) )
                break;
            else
                continue;
        }

        addrlen = sizeof(cliaddr);
        new_sockfd = accept( srv.lsockfd, (struct sockaddr*)&cliaddr, &addrlen );
        if( -1 == new_sockfd ) {
            mperror( g_flog, errno,  "accept" );
            break;
        }
        TRACE( (void)fprintf( g_flog, "Accepted socket=[%d]\n",
                    new_sockfd) );

        /* service client's request:
         *   should a non-critical error occur, bail out on the client
         *   but do not stop the server
         */
        do {
            rc = set_timeouts(new_sockfd, new_sockfd, SRVSOCK_TIMEOUT, SRVSOCK_TIMEOUT);
            if( 0 != rc ) break;

            rc = read_command( new_sockfd, srv.cmd, sizeof(srv.cmd), param, sizeof(param) );
            if( 0 != rc ) break;

            rc = process_command( new_sockfd, &srv, param, sizeof(param) );
        }
        while(0);

        (void) close( new_sockfd );
        TRACE( (void)fprintf( g_flog, "Closed accepted socket [%d]\n", new_sockfd) );
        new_sockfd = -1;

    } /* server loop */


    TRACE( (void)fprintf( g_flog, "Exited server loop\n") );

    if( new_sockfd > 0 ) {
        (void) close( new_sockfd );
    }

    wait_terminated( &srv );
    terminate_all_clients( &srv );
    wait_all( &srv );

    (void) close( srv.lsockfd );

    free_server_ctx( &srv );

    (void)fprintf( g_flog, "Server [%d] exits rc=[%d]\n", getpid(), rc );
    return rc;
}


static void
usage( const char* app, FILE* fp )
{
    (void) fprintf(fp, "%s %s (build %d) %s\n", app, VERSION, BUILDNUM,
            COMPILE_MODE );
    (void) fprintf(fp, "usage: %s [-vTS] [-a listenaddr] -p port "
            "[-m mcast_ifc_addr] [-c clients] [-l logfile] "
            "[-B sizeK] [-n nice_incr]\n", app );
    (void) fprintf(fp,
            "\t-v : enable verbose output [default = disabled]\n"
            "\t-S : enable client statistics [default = disabled]\n"
            "\t-T : do NOT run as a daemon [default = daemon if root]\n"
            "\t-a : (IPv4) address/interface to listen on [default = %s]\n"
            "\t-p : port to listen on\n"
            "\t-m : (IPv4) address/interface of (multicast) source [default = %s]\n"
            "\t-c : max clients to serve [default = %d, max = %d]\n"
            "\t-l : log output to file [default = stderr]\n"
            "\t-B : cache size for inbound (multicast) data [default = %uK]\n"
            "\t-n : nice value increment [default = %d]\n",
            IPv4_ALL, IPv4_ALL, DEFAULT_CLIENT_COUNT, MAX_CLIENT_COUNT,
            DEFAULT_CACHE_K, g_uopt.nice_incr );
    (void) fprintf( fp, "Examples:\n"
            "  %s -p 4022 \n"
            "\tlisten for HTTP requests on port 4022, all network interfaces\n"
            "  %s -a lan0 -p 4022 -m lan1\n"
            "\tlisten for HTTP requests on interface lan0, port 4022;\n"
            "\tsubscribe to multicast groups on interface lan1\n",
            app, app);
    (void) fprintf( fp, "\n  %s\n\n", COPYRIGHT_NOTICE );
    return;
}


static void
set_verbose()
{
#ifdef TRACE_MODULE
    g_uopt.is_verbose = uf_TRUE;
#else
    (void) fputs("Verbose info has not been compiled in.\n", stderr);
    exit(ERR_PARAM);
#endif
}


int
main( int argc, char* const argv[] )
{
    int rc, ch, port, custom_log, no_daemon;
    char ipaddr[IPADDR_STR_SIZE],
         mcast_addr[IPADDR_STR_SIZE];

    char pidfile[ MAXPATHLEN ];
    struct sigaction qact, cact, oldact;

    rc = 0;
    ipaddr[0] = mcast_addr[0] = pidfile[0] = '\0';
    port = -1;
    custom_log = no_daemon = 0;

    init_uopt( &g_uopt );

    while( -1 != (ch = getopt(argc, argv, "TvSa:l:p:m:c:B:n:R:r:w:")) ) {
        switch( ch ) {
            case 'v': set_verbose();    break;
            case 'T': no_daemon = 1;    break;
            case 'S': g_uopt.cl_tpstat = uf_TRUE;
                                        break;
            case 'a':
                      rc = get_ipv4_address( optarg, ipaddr, sizeof(ipaddr) );
                      if( 0 != rc ) {
                        (void) fprintf( stderr, "Invalid address: [%s]\n",
                                        optarg );
                          exit(ERR_PARAM);
                      }
                      break;

            case 'p':
                      port = atoi( optarg );
                      if( port <= 0 ) {
                        (void) fprintf( stderr, "Invalid port number: [%d]\n",
                                        port );
                        exit(ERR_PARAM);
                      }
                      break;

            case 'm':
                      rc = get_ipv4_address( optarg, mcast_addr, sizeof(mcast_addr) );
                      if( 0 != rc ) {
                        (void) fprintf( stderr, "Invalid multicast address: [%s]\n",
                                        optarg );
                          exit(ERR_PARAM);
                      }
                      break;

            case 'c':
                      g_uopt.max_clients = atoi( optarg );
                      if( (g_uopt.max_clients < MIN_CLIENT_COUNT) ||
                          (g_uopt.max_clients > MAX_CLIENT_COUNT) ) {
                        (void) fprintf( stderr,
                                "Client count should be between %d and %d\n",
                                MIN_CLIENT_COUNT, MAX_CLIENT_COUNT );
                        exit(ERR_PARAM);
                      }
                      break;

            case 'l':
                      g_flog = fopen( optarg, "a" );
                      if( NULL == g_flog ) {
                        rc = errno;
                        (void) fprintf( stderr, "Error opening logfile [%s]: %s\n",
                                optarg, strerror(rc) );
                        exit( ERR_PARAM );
                      }

                      Setlinebuf( g_flog );
                      custom_log = 1;
                      break;

            case 'B':
                      g_uopt.rbuf_K = atoi(optarg);
                      if( (g_uopt.rbuf_K < MIN_MCACHE_K) ||
                          (g_uopt.rbuf_K > MAX_MCACHE_K) ) {
                        fprintf(stderr, "Multicast socket buffer size "
                                "must be within [%u-%u] Kb\n",
                                MIN_MCACHE_K, MAX_MCACHE_K );
                        exit(ERR_PARAM);
                      }
                      break;

            case 'n':
                      g_uopt.nice_incr = atoi( optarg );
                      if( 0 == g_uopt.nice_incr ) {
                        (void) fprintf( stderr,
                            "Invalid nice-value increment: [%s]\n", optarg );
                        exit(ERR_PARAM);
                      }
                      break;

            case 'R':
                      g_uopt.rbuf_msgs = atoi( optarg );
                      if( (g_uopt.rbuf_msgs <= 0) && (-1 != g_uopt.rbuf_msgs) ) {
                        (void) fprintf( stderr,
                                "Invalid rcache size: [%s]\n", optarg );
                        exit(ERR_PARAM);
                      }
                      break;

            case 'r':
                      if( 0 != access(optarg, R_OK) ) {
                        perror("source file - access");
                        exit(ERR_PARAM);
                      }

                      g_uopt.srcfile = strdup( optarg );
                      break;

            case 'w':
                      g_uopt.dstfile = strdup( optarg );
                      break;

            default:
                     usage( argv[0], stderr );
                     exit(ERR_PARAM);
        }
    } /* while getopt */

    if( '\0' == mcast_addr[0] ) {
        (void) strncpy( mcast_addr, IPv4_ALL, sizeof(mcast_addr) - 1 );
    }

    if( (argc < 2) || (port <= 0) ) {
        usage( argv[0], stderr );
        exit(ERR_PARAM);
    }

    if( !custom_log ) {
        /* in debug mode output goes to stderr, otherwise to /dev/null */
        g_flog = ((uf_TRUE == g_uopt.is_verbose)
                ? stderr
                : fopen( "/dev/null", "a" ));
        if( NULL == g_flog ) {
            perror("fopen");
            exit(ERR_INTERNAL);
        }
    }

    openlog( g_appname, LOG_CONS | LOG_PID, LOG_LOCAL0 );

    if( 0 == geteuid() ) {
        rc = set_pidfile( g_appname, pidfile, sizeof(pidfile) );
        if( 0 != rc ) {
            perror("set_pidfile");
            exit(ERR_INTERNAL);
        }

        if( !no_daemon ) {
            if( 0 != (rc = daemonize(0)) )
                exit(ERR_INTERNAL);
        }

        rc = make_pidfile( pidfile, getpid(), g_flog );
        if( 0 != rc ) exit(ERR_INTERNAL);
    }

    qact.sa_handler = handle_quitsigs;
    sigemptyset(&qact.sa_mask);
    qact.sa_flags = 0;

    if( (sigaction(SIGTERM, &qact, &oldact) < 0) ||
        (sigaction(SIGQUIT, &qact, &oldact) < 0) ||
        (sigaction(SIGINT,  &qact, &oldact) < 0) ||
        (sigaction(SIGPIPE, &qact, &oldact) < 0)) {
        perror("sigaction-quit");
        exit(ERR_INTERNAL);
    }

    cact.sa_handler = handle_sigchld;
    sigemptyset(&cact.sa_mask);
    cact.sa_flags = 0;

    if( sigaction(SIGCHLD, &cact, &oldact) < 0 ) {
        perror("sigaction-sigchld");
        exit(ERR_INTERNAL);
    }

    syslog( LOG_NOTICE, "%s is starting\n", g_appname );

    rc = server_loop( ipaddr, port, mcast_addr );

    syslog( LOG_NOTICE, "%s is exiting rc=[%d]\n",
            g_appname, rc);

    if( '\0' != pidfile[0] ) {
        if( -1 == unlink(pidfile) ) {
            mperror( g_flog, errno, "unlink" );
        }
    }

    if( g_flog && (stderr != g_flog) ) {
        (void) fclose(g_flog);
    }

    closelog();
    free_uopt( &g_uopt );

    return rc;
}

/* __EOF__ */

