/* @(#) implementation of utility functions for udpxy
 *
 * Copyright 2008 Pavel V. Cherenkov
 *
 *  This file is part of udpxy.
 *
 *  udpxy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  udpxy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with udpxy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>

#include "util.h"
#include "mtrace.h"
#include "osdef.h"

/* write buffer to a file
 *
 */
ssize_t
save_buffer( const void* buf, size_t len, const char* filename )
{
    int fd, rc;
    ssize_t nw, left;
    const char* p;

    assert( buf && len && filename );

    rc = 0;
    fd = creat( filename,
            (mode_t)(S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) );
    if( -1 == fd ) {
        perror("creat");
        return 1;
    }

    for( p = (const char*)buf, nw = 0, left = len; left > 0; ) {
        nw = write( fd, p, left );
        if( nw <= 0 ) {
            nw = 0;
            if( EINTR != errno) {
                perror("write");
                rc = -1;
                break;
            }
        }

        left -= nw;
        p += nw;
    }

    (void)close(fd);
    return (rc ? -1 : ((ssize_t)len - left));
}


/* make current process run as a daemon
 */
int
daemonize(int options)
{
    pid_t pid;
    int rc, fh;

    rc = fh = 0;

    if( (pid = fork()) < 0 ) {
        perror("fork");
        return -1;
    }
    else if( (pid_t)0 != pid ) {
        exit(0);
    }

    do {
        if( -1 == (rc = setsid()) ) {
            perror("setsid");
            break;
        }

        if( -1 == (rc = chdir("/")) )
            break;
        (void) umask(0);

        if( !(options & DZ_STDIO_OPEN) ) {
            for( fh = 0; fh < 3; ++fh )
                if( -1 == (rc = close(fh)) ) {
                    perror("close");
                    break;
                }
        }

        if( SIG_ERR == signal(SIGHUP, SIG_IGN) ) {
            rc = 2;
            break;
        }

    } while(0);

    if( 0 != rc ) return rc;

    /* child exits to avoid session leader's re-acquiring
     * control terminal */
    if( (pid = fork()) < 0 ) {
        perror("fork");
        return -1;
    }
    else if( (pid_t)0 != pid )
        exit(0);

    return 0;
}


/* multiplex error output to custom log
 * and syslog
 */
void
mperror( FILE* fp, int err, const char* str )
{
    char buf[ 256 ];

    assert(str);

    snprintf( buf, sizeof(buf) - 1, "%s: %s",
            str, strerror(err) );
    buf[ sizeof(buf) - 1 ] = '\0';

    syslog( LOG_ERR | LOG_LOCAL0, "%s", buf );
    if( fp ) (void) fprintf( fp, "%s\n", buf );

    return;
}


/* write-lock on a file handle
 */
static int
wlock_file( int fd )
{
    struct flock  lck;

    lck.l_type      = F_WRLCK;
    lck.l_start     = 0;
    lck.l_whence    = SEEK_SET;
    lck.l_len       = 0;

    return fcntl( fd, F_SETLK, &lck );
}


/* create and lock file with process's ID
 */
int
make_pidfile( const char* fpath, pid_t pid, FILE* log )
{
    int fd = -1, rc = 0, n = -1;
    ssize_t nwr = -1;

    #define LLONG_MAX_DIGITS 21
    char buf[ LLONG_MAX_DIGITS + 1 ];

    mode_t fmode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;

    assert( (NULL != fpath) && pid );

    errno = 0;
    do {
        fd = open( fpath, O_CREAT | O_WRONLY | O_NOCTTY, fmode );
        if( -1 == fd ) {
            mperror(log, errno, "make_pidfile - open");
            rc = EXIT_FAILURE;
            break;
        }

        rc = wlock_file( fd );
        if( 0 != rc ) {
            if( (EACCES == errno) || (EAGAIN == errno) ) {
                (void) fprintf( stderr, "File [%s] is locked "
                    "(another instance of daemon must be running)\n",
                    fpath );
                exit(EXIT_FAILURE);
            }

            mperror(log, errno, "wlock_file");
            break;
        }

        rc = ftruncate( fd, 0 );
        if( 0 != rc ) {
            mperror(log, errno, "make_pidfile - ftruncate");
            break;
        }

        n = snprintf( buf, sizeof(buf) - 1, "%d", pid );
        if( n < 0 ) {
            mperror(log, errno, "make_pidfile - snprintf");
            rc = EXIT_FAILURE;
            break;
        }

        nwr = write( fd, buf, n );
        if( (ssize_t)n != nwr ) {
            mperror( log, errno, "make_pidfile - write");
            rc = EXIT_FAILURE;
            break;
        }

    } while(0);

    if( (0 != rc) && (fd > 0) ) {
        (void)close( fd );
    }

    return rc;
}


/* write path to application's pidfile into the the buffer
 * (fail if destination directory is not writable)
 */
int
set_pidfile( const char* appname, char* buf, size_t len )
{
    int n = -1;

    assert( appname && buf && len );

    if( -1 == access(PIDFILE_DIR, W_OK ) )
        return -1;

    n = snprintf( buf, len, "%s/%s.pid", PIDFILE_DIR, appname );
    if( n < 0 ) return EXIT_FAILURE;

    buf[ len - 1 ] = '\0';

    return 0;
}


/* write buffer to designated socket/file
 */
ssize_t
write_buf( int fd, const char* data, const ssize_t len, FILE* log )
{
    ssize_t n = 0, nwr = 0;

    for( n = 0; errno = 0, n < len ; ) {
        nwr = write( fd, &(data[n]), len - n );
        if( nwr <= 0 ) {
            if( EINTR == errno ) {
                TRACE( (void)fprintf( log, "%s interrupted\n", __func__ ) );
                continue;
            }
            else
                break;
        }

        if( nwr != len ) {
            if( NULL != log ) {
                TRACE( (void)fprintf( log, "Fragment written [%ld]/[%ld] bytes\n",
                        (long)nwr, (long)len ) );
            }
        }

        n += nwr;
    }

    if( nwr <= 0 ) {
        if( log )
            mperror( log, errno, "write_buf - write" );
        return -1;
    }

    return n;
}


/* read data chunk of designated size (or less) into buffer
 * (will *NOT* attempt to re-read if read less than expected
 *  w/o interruption)
 */
ssize_t
read_buf( int fd, char* data, const ssize_t len, FILE* log )
{
    ssize_t n = 0, nrd = 0;

    for( n = 0; errno = 0, n < len ; ) {
        nrd = read( fd, &(data[n]), len - n );
        if( nrd <= 0 ) {
            if( EINTR == errno ) {
                TRACE( (void)fprintf( log, "%s interrupted\n", __func__ ) );
                errno = 0;
                continue;
            }
            else
                break;
        }

        n += nrd;
        /*
        if( nrd != len ) {
            if( NULL != log ) {
                TRACE( (void)fprintf( log, "Fragment read [%ld]/[%ld] bytes\n",
                        (long)nrd, (long)len ) );
            }
        }
        */

        /* we only read as much as we can read on once (uninterrupted) */
        break;
    }

    if( nrd < 0 ) {
        if( log ) mperror( log, errno, "read" );
    }

    return n;
}


/* output hex dump of a memory fragment
 */
void
hex_dump( const char* msg, const char* data, size_t len, FILE* log )
{
    u_int i = 0;

    assert( data && (len > 0) && log );

    if( msg ) fprintf( log, "%s: ", msg );

    for( i = 0; i < len; ++i )
        (void)fprintf( log, "%02x ", data[i] & 0xFF );

    (void) fputs("\n", log);
    return;
}


/* check for expected size, complain if not matched
 */
int
sizecheck( const char* msg, ssize_t expct, ssize_t len,
              FILE* log, const char* func )
{
    assert( msg && log && func );
    if( expct == len ) return 0;

    TRACE( (void)fprintf( log, "%s: %s - read only [%ld] "
                "bytes out of [%ld]\n",
                func, msg, (long)len, (long)expct ) );

    return 1;
}


/* check for a potential buffer overrun by
 * evaluating target buffer and the portion
 * of that buffer to be accessed
 */
int
buf_overrun( const char* buf, size_t buflen,
             size_t offset, size_t dlen,
             FILE* log )
{
    const char* tgt = buf + offset + dlen;
    if( tgt > (buf + buflen) ) {
        if( NULL != log ) {
            TRACE( (void)fprintf( log, "+++ BUFFER OVERRUN at: "
                    "buf=[%p], size=[%lu] - intending to access [%p]: "
                    "offset=[%lu], data_length=[%lu]\n",
                    buf, (u_long)buflen, tgt, (u_long)offset, (u_long)dlen) );
        }
        return 1;
    }

    return 0;
}


/* __EOF__ */

